# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class AdminFacade(Type):
    name = "Admin"
    version = 3

    @ReturnMapping(LoginResult)
    async def Login(
        self,
        auth_tag=None,
        bakery_version=None,
        cli_args=None,
        client_version=None,
        credentials=None,
        macaroons=None,
        nonce=None,
        token=None,
        user_data=None,
    ):
        """Login logs in with the provided credentials.  All subsequent requests on the
        connection will act as the authenticated user.

        auth_tag : str
        bakery_version : int
        cli_args : str
        client_version : str
        credentials : str
        macaroons : typing.Sequence[~Macaroon]
        nonce : str
        token : str
        user_data : str
        Returns -> LoginResult
        """
        if auth_tag is not None and not isinstance(auth_tag, (bytes, str)):
            raise Exception(
                f"Expected auth_tag to be a str, received: {type(auth_tag)}"
            )

        if bakery_version is not None and not isinstance(bakery_version, int):
            raise Exception(
                f"Expected bakery_version to be a int, received: {type(bakery_version)}"
            )

        if cli_args is not None and not isinstance(cli_args, (bytes, str)):
            raise Exception(
                f"Expected cli_args to be a str, received: {type(cli_args)}"
            )

        if client_version is not None and not isinstance(client_version, (bytes, str)):
            raise Exception(
                f"Expected client_version to be a str, received: {type(client_version)}"
            )

        if credentials is not None and not isinstance(credentials, (bytes, str)):
            raise Exception(
                f"Expected credentials to be a str, received: {type(credentials)}"
            )

        if macaroons is not None and not isinstance(macaroons, (bytes, str, list)):
            raise Exception(
                f"Expected macaroons to be a Sequence, received: {type(macaroons)}"
            )

        if nonce is not None and not isinstance(nonce, (bytes, str)):
            raise Exception(f"Expected nonce to be a str, received: {type(nonce)}")

        if token is not None and not isinstance(token, (bytes, str)):
            raise Exception(f"Expected token to be a str, received: {type(token)}")

        if user_data is not None and not isinstance(user_data, (bytes, str)):
            raise Exception(
                f"Expected user_data to be a str, received: {type(user_data)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Admin", request="Login", version=3, params=_params)
        _params["auth-tag"] = auth_tag
        _params["bakery-version"] = bakery_version
        _params["cli-args"] = cli_args
        _params["client-version"] = client_version
        _params["credentials"] = credentials
        _params["macaroons"] = macaroons
        _params["nonce"] = nonce
        _params["token"] = token
        _params["user-data"] = user_data
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(RedirectInfoResult)
    async def RedirectInfo(self):
        """RedirectInfo returns redirected host information for the model.
        In Juju it always returns an error because the Juju controller
        does not multiplex controllers.

        Returns -> RedirectInfoResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Admin", request="RedirectInfo", version=3, params=_params)

        reply = await self.rpc(msg)
        return reply


class AllWatcherFacade(Type):
    name = "AllWatcher"
    version = 3

    @ReturnMapping(AllWatcherNextResults)
    async def Next(self):
        """Next will return the current state of everything on the first call
        and subsequent calls will

        Returns -> AllWatcherNextResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="AllWatcher", request="Next", version=3, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def Stop(self):
        """Stop stops the watcher.

        Returns -> None
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="AllWatcher", request="Stop", version=3, params=_params)

        reply = await self.rpc(msg)
        return reply

    async def rpc(self, msg):
        """Patch rpc method to add Id."""
        if not hasattr(self, "Id"):
            raise RuntimeError('Missing "Id" field')
        msg["Id"] = id

        from .facade import TypeEncoder

        reply = await self.connection.rpc(msg, encoder=TypeEncoder)
        return reply


class BackupsFacade(Type):
    name = "Backups"
    version = 3

    @ReturnMapping(BackupsMetadataResult)
    async def Create(self, no_download=None, notes=None):
        """Create is the API method that requests juju to create a new backup
        of its state.

        no_download : bool
        notes : str
        Returns -> BackupsMetadataResult
        """
        if no_download is not None and not isinstance(no_download, bool):
            raise Exception(
                f"Expected no_download to be a bool, received: {type(no_download)}"
            )

        if notes is not None and not isinstance(notes, (bytes, str)):
            raise Exception(f"Expected notes to be a str, received: {type(notes)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Backups", request="Create", version=3, params=_params)
        _params["no-download"] = no_download
        _params["notes"] = notes
        reply = await self.rpc(msg)
        return reply


class ModelConfigFacade(Type):
    name = "ModelConfig"
    version = 3

    @ReturnMapping(GetConstraintsResults)
    async def GetModelConstraints(self):
        """GetModelConstraints returns the constraints for the model.

        Returns -> GetConstraintsResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelConfig", request="GetModelConstraints", version=3, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelConfigResults)
    async def ModelGet(self):
        """ModelGet implements the server-side part of the
        model-config CLI command.

        Returns -> ModelConfigResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelConfig", request="ModelGet", version=3, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def ModelSet(self, config=None):
        """ModelSet implements the server-side part of the
        set-model-config CLI command.

        config : typing.Mapping[str, typing.Any]
        Returns -> None
        """
        if config is not None and not isinstance(config, dict):
            raise Exception(
                f"Expected config to be a Mapping, received: {type(config)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelConfig", request="ModelSet", version=3, params=_params)
        _params["config"] = config
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def ModelUnset(self, keys=None):
        """ModelUnset implements the server-side part of the
        set-model-config CLI command.

        keys : typing.Sequence[str]
        Returns -> None
        """
        if keys is not None and not isinstance(keys, (bytes, str, list)):
            raise Exception(f"Expected keys to be a Sequence, received: {type(keys)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelConfig", request="ModelUnset", version=3, params=_params)
        _params["keys"] = keys
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringResult)
    async def SLALevel(self):
        """SLALevel returns the current sla level for the model.

        Returns -> StringResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelConfig", request="SLALevel", version=3, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelSequencesResult)
    async def Sequences(self):
        """Sequences returns the model's sequence names and next values.

        Returns -> ModelSequencesResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelConfig", request="Sequences", version=3, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def SetModelConstraints(self, application=None, constraints=None):
        """SetModelConstraints sets the constraints for the model.

        application : str
        constraints : Value
        Returns -> None
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if constraints is not None and not isinstance(constraints, (dict, Value)):
            raise Exception(
                f"Expected constraints to be a Value, received: {type(constraints)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelConfig", request="SetModelConstraints", version=3, params=_params
        )
        _params["application"] = application
        _params["constraints"] = constraints
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def SetSLALevel(self, modelslainfo=None, creds=None, level=None, owner=None):
        """SetSLALevel sets the sla level on the model.

        modelslainfo : ModelSLAInfo
        creds : typing.Sequence[int]
        level : str
        owner : str
        Returns -> None
        """
        if modelslainfo is not None and not isinstance(
            modelslainfo, (dict, ModelSLAInfo)
        ):
            raise Exception(
                f"Expected modelslainfo to be a ModelSLAInfo, received: {type(modelslainfo)}"
            )

        if creds is not None and not isinstance(creds, (bytes, str, list)):
            raise Exception(f"Expected creds to be a Sequence, received: {type(creds)}")

        if level is not None and not isinstance(level, (bytes, str)):
            raise Exception(f"Expected level to be a str, received: {type(level)}")

        if owner is not None and not isinstance(owner, (bytes, str)):
            raise Exception(f"Expected owner to be a str, received: {type(owner)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelConfig", request="SetSLALevel", version=3, params=_params)
        _params["ModelSLAInfo"] = modelslainfo
        _params["creds"] = creds
        _params["level"] = level
        _params["owner"] = owner
        reply = await self.rpc(msg)
        return reply


class ResourcesFacade(Type):
    name = "Resources"
    version = 3

    @ReturnMapping(AddPendingResourcesResult)
    async def AddPendingResources(
        self,
        entity=None,
        charm_origin=None,
        macaroon=None,
        resources=None,
        tag=None,
        url=None,
    ):
        """AddPendingResources adds the provided resources (info) to the Juju
        model in a pending state, meaning they are not available until
        resolved. Handles CharmHub and Local charms.

        entity : Entity
        charm_origin : CharmOrigin
        macaroon : Macaroon
        resources : typing.Sequence[~CharmResource]
        tag : str
        url : str
        Returns -> AddPendingResourcesResult
        """
        if entity is not None and not isinstance(entity, (dict, Entity)):
            raise Exception(f"Expected entity to be a Entity, received: {type(entity)}")

        if charm_origin is not None and not isinstance(
            charm_origin, (dict, CharmOrigin)
        ):
            raise Exception(
                f"Expected charm_origin to be a CharmOrigin, received: {type(charm_origin)}"
            )

        if macaroon is not None and not isinstance(macaroon, (dict, Macaroon)):
            raise Exception(
                f"Expected macaroon to be a Macaroon, received: {type(macaroon)}"
            )

        if resources is not None and not isinstance(resources, (bytes, str, list)):
            raise Exception(
                f"Expected resources to be a Sequence, received: {type(resources)}"
            )

        if tag is not None and not isinstance(tag, (bytes, str)):
            raise Exception(f"Expected tag to be a str, received: {type(tag)}")

        if url is not None and not isinstance(url, (bytes, str)):
            raise Exception(f"Expected url to be a str, received: {type(url)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Resources", request="AddPendingResources", version=3, params=_params
        )
        _params["Entity"] = entity
        _params["charm-origin"] = charm_origin
        _params["macaroon"] = macaroon
        _params["resources"] = resources
        _params["tag"] = tag
        _params["url"] = url
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ResourcesResults)
    async def ListResources(self, entities=None):
        """ListResources returns the list of resources for the given application.

        entities : typing.Sequence[~Entity]
        Returns -> ResourcesResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Resources", request="ListResources", version=3, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply


class UserManagerFacade(Type):
    name = "UserManager"
    version = 3

    @ReturnMapping(AddUserResults)
    async def AddUser(self, users=None):
        """AddUser adds a user with a username, and either a password or
        a randomly generated secret key which will be returned.

        users : typing.Sequence[~AddUser]
        Returns -> AddUserResults
        """
        if users is not None and not isinstance(users, (bytes, str, list)):
            raise Exception(f"Expected users to be a Sequence, received: {type(users)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="UserManager", request="AddUser", version=3, params=_params)
        _params["users"] = users
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def DisableUser(self, entities=None):
        """DisableUser disables one or more users.  If the user is already disabled,
        the action is considered a success.

        entities : typing.Sequence[~Entity]
        Returns -> ErrorResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="UserManager", request="DisableUser", version=3, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def EnableUser(self, entities=None):
        """EnableUser enables one or more users.  If the user is already enabled,
        the action is considered a success.

        entities : typing.Sequence[~Entity]
        Returns -> ErrorResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="UserManager", request="EnableUser", version=3, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelUserInfoResults)
    async def ModelUserInfo(self, entities=None):
        """ModelUserInfo returns information on all users in the model.

        entities : typing.Sequence[~Entity]
        Returns -> ModelUserInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="UserManager", request="ModelUserInfo", version=3, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def RemoveUser(self, entities=None):
        """RemoveUser permanently removes a user from the current controller for each
        entity provided. While the user is permanently removed we keep it's
        information around for auditing purposes.
        TODO(redir): Add information about getting deleted user information when we
        add that capability.

        entities : typing.Sequence[~Entity]
        Returns -> ErrorResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="UserManager", request="RemoveUser", version=3, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(AddUserResults)
    async def ResetPassword(self, entities=None):
        """ResetPassword resets password for supplied users by
        invalidating current passwords (if any) and generating
        new random secret keys which will be returned.
        Users cannot reset their own password.

        entities : typing.Sequence[~Entity]
        Returns -> AddUserResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="UserManager", request="ResetPassword", version=3, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def SetPassword(self, changes=None):
        """SetPassword changes the stored password for the specified users.

        changes : typing.Sequence[~EntityPassword]
        Returns -> ErrorResults
        """
        if changes is not None and not isinstance(changes, (bytes, str, list)):
            raise Exception(
                f"Expected changes to be a Sequence, received: {type(changes)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="UserManager", request="SetPassword", version=3, params=_params)
        _params["changes"] = changes
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(UserInfoResults)
    async def UserInfo(self, entities=None, include_disabled=None):
        """UserInfo returns information on a user.

        entities : typing.Sequence[~Entity]
        include_disabled : bool
        Returns -> UserInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        if include_disabled is not None and not isinstance(include_disabled, bool):
            raise Exception(
                f"Expected include_disabled to be a bool, received: {type(include_disabled)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="UserManager", request="UserInfo", version=3, params=_params)
        _params["entities"] = entities
        _params["include-disabled"] = include_disabled
        reply = await self.rpc(msg)
        return reply



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
from __future__ import annotations

import asyncio
import base64
import json
import logging
import ssl
import urllib.request
import warnings
import weakref
from http.client import HTTPSConnection
from typing import Any, Literal, Sequence

import macaroonbakery.bakery as bakery
import macaroonbakery.httpbakery as httpbakery
import websockets
from dateutil.parser import parse
from typing_extensions import Self, TypeAlias, overload

from juju import errors, tag, utils
from juju.client import client
from juju.utils import IdQueue
from juju.version import CLIENT_VERSION

from .facade import TypeEncoder, _Json, _RichJson
from .facade_versions import client_facade_versions, known_unsupported_facades

SpecifiedFacades: TypeAlias = "dict[str, dict[Literal['versions'], Sequence[int]]]"
_WebSocket: TypeAlias = websockets.WebSocketClientProtocol

LEVELS = ["TRACE", "DEBUG", "INFO", "WARNING", "ERROR"]
log = logging.getLogger("juju.client.connection")


class Monitor:
    """Monitor helper class for our Connection class.

    Contains a reference to an instantiated Connection, along with a
    reference to the Connection.receiver Future. Upon inspection of
    these objects, this class determines whether the connection is in
    an 'error', 'connected' or 'disconnected' state.

    Use this class to stay up to date on the health of a connection,
    and take appropriate action if the connection errors out due to
    network issues or other unexpected circumstances.

    """

    ERROR = "error"
    CONNECTED = "connected"
    DISCONNECTING = "disconnecting"
    DISCONNECTED = "disconnected"

    def __init__(self, connection: Connection):
        self.connection = weakref.ref(connection)
        self.reconnecting = asyncio.Lock()
        self.close_called = asyncio.Event()

    @property
    def status(self):
        """Determine the status of the connection and receiver, and return
        ERROR, CONNECTED, or DISCONNECTED as appropriate.

        For simplicity, we only consider ourselves to be connected
        after the Connection class has setup a receiver task. This
        only happens after the websocket is open, and the connection
        isn't usable until that receiver has been started.

        """
        connection = self.connection()

        # the connection instance was destroyed but someone kept
        # a separate reference to the monitor for some reason
        if not connection:
            return self.DISCONNECTED

        # connection cleanly disconnected or not yet opened
        if not connection._ws:
            return self.DISCONNECTED

        # close called but not yet complete
        if self.close_called.is_set():
            return self.DISCONNECTING

        # connection closed uncleanly (we didn't call connection.close)
        if connection.is_debug_log_connection:
            stopped = connection._debug_log_task.cancelled()
        else:
            stopped = (
                connection._receiver_task is not None
                and connection._receiver_task.cancelled()
            )

        if stopped or connection._ws.state is not websockets.protocol.State.OPEN:
            return self.ERROR

        # everything is fine!
        return self.CONNECTED


class Connection:
    """Usage::

    # Connect to an arbitrary api server
    client = await Connection.connect(
        api_endpoint, model_uuid, username, password, cacert)

    """

    # Juju 3.6 is limited by MongoDB transaction size limit, 16MB of BSON.
    # Set the jRPC receive limit to 64MB of JSON to account for formatting differences.
    MAX_FRAME_SIZE = 0x4_000_000
    "Maximum size for a single frame, 64MB."
    facades: dict[str, int]
    _specified_facades: dict[str, Sequence[int]]
    bakery_client: Any
    usertag: str | None
    password: str | None
    name: str
    __request_id__: int
    endpoints: list[tuple[str, str]] | None  # Set by juju/controller.py
    is_debug_log_connection: bool
    monitor: Monitor
    proxy: Any  # Need to find types for this library
    max_frame_size: int
    _retries: int
    _retry_backoff: float
    uuid: str | None
    messages: IdQueue
    _ws: _WebSocket | None

    @classmethod
    async def connect(
        cls,
        endpoint=None,
        uuid: str | None = None,
        username: str | None = None,
        password: str | None = None,
        cacert=None,
        bakery_client=None,
        max_frame_size: int | None = None,
        retries=3,
        retry_backoff=10,
        specified_facades: SpecifiedFacades | None = None,
        proxy=None,
        debug_log_conn=None,
        debug_log_params={},
    ) -> Self:
        """Connect to the websocket.

        If uuid is None, the connection will be to the controller. Otherwise it
        will be to the model.

        :param str endpoint: The hostname:port of the controller to connect to (or list of strings).
        :param str uuid: The model UUID to connect to (None for a
            controller-only connection).
        :param str username: The username for controller-local users (or None
            to use macaroon-based login.)
        :param str password: The password for controller-local users.
        :param str cacert: The CA certificate of the controller
            (PEM formatted).
        :param httpbakery.Client bakery_client: The macaroon bakery client to
            to use when performing macaroon-based login. Macaroon tokens
            acquired when logging will be saved to bakery_client.cookies.
            If this is None, a default bakery_client will be used.
        :param int max_frame_size: The maximum websocket frame size to allow.
        :param int retries: When connecting or reconnecting, and all endpoints
            fail, how many times to retry the connection before giving up.
        :param int retry_backoff: Number of seconds to increase the wait
            between connection retry attempts (a backoff of 10 with 3 retries
            would wait 10s, 20s, and 30s).
        :param specified_facades: (deprecated) define a series of facade versions you wish to override
            to prevent using the conservative client pinning with in the client.
        :param TextIOWrapper debug_log_conn: target if this is a debug log connection
        :param dict debug_log_params: filtering parameters for the debug-log output
        """
        self = cls()
        if endpoint is None:
            raise ValueError("no endpoint provided")
        if not isinstance(endpoint, str) and not isinstance(endpoint, list):
            raise TypeError("Endpoint should be either str or list")
        self.uuid = uuid
        if bakery_client is None:
            bakery_client = httpbakery.Client()
        self.bakery_client = bakery_client
        if username and "@" in username and not username.endswith("@local"):
            # We're trying to log in as an external user - we need to use
            # macaroon authentication with no username or password.
            if password is not None:
                raise errors.JujuAuthError(
                    "cannot log in as external user with a password"
                )
            username = None
        self.usertag = tag.user(username)
        self.password = password

        self.__request_id__ = 0

        # The following instance variables are initialized by the
        # _connect_with_redirect method, but create them here
        # as a reminder that they will exist.
        self.addr = None
        self._ws = None
        self.endpoint = None
        self.endpoints = None
        self.cacert = None
        self.info = None

        self.debug_log_target = debug_log_conn
        self.is_debug_log_connection = debug_log_conn is not None
        self.debug_log_params = debug_log_params
        self.debug_log_shown_lines = 0  # number of lines

        # Create that _Task objects but don't start the tasks yet.
        self._pinger_task = None
        self._receiver_task = None
        self._debug_log_task = None

        self._retries = retries
        self._retry_backoff = retry_backoff

        self.facades = {}

        if specified_facades:
            warnings.warn(
                "The `specified_facades` argument is deprecated and will be removed soon",
                DeprecationWarning,
                stacklevel=3,
            )
            self._specified_facades = {
                name: d["versions"] for name, d in specified_facades.items()
            }
        else:
            self._specified_facades = {}

        self.messages = IdQueue()
        self.monitor = Monitor(connection=self)
        if max_frame_size is None:
            max_frame_size = self.MAX_FRAME_SIZE
        self.max_frame_size = max_frame_size

        self.proxy = proxy
        if self.proxy is not None:
            self.proxy.connect()

        _endpoints = (
            [(endpoint, cacert)]
            if isinstance(endpoint, str)
            else [(e, cacert) for e in endpoint]
        )
        last_error = None
        for _ep in _endpoints:
            try:
                if self.is_debug_log_connection:
                    # make a direct connection with basic auth if
                    # debug-log (i.e. no redirection or login)
                    await self._connect([_ep])
                else:
                    await self._connect_with_redirect([_ep])
                return self
            except ssl.SSLError as e:
                last_error = e
                continue
            except OSError as e:
                logging.debug(f"Cannot access endpoint {_ep}: {e.strerror}")
                last_error = e
                continue
        if last_error is not None:
            raise last_error
        raise Exception("Unable to connect to websocket")

    @property
    def ws(self):
        log.warning(
            "Direct access to the websocket object may cause disruptions in asyncio event handling."
        )
        return self._ws

    @property
    def username(self) -> str | None:
        if not self.usertag:
            return None
        return self.usertag[len("user-") :]

    @property
    def is_using_old_client(self):
        if self.info is None:
            raise errors.JujuError("Not connected yet.")
        return self.info["server-version"].startswith("2.")

    @property
    def is_open(self):
        return self.monitor.status == Monitor.CONNECTED

    def _get_ssl(self, cert: str | None = None) -> ssl.SSLContext:
        context = ssl.create_default_context(
            purpose=ssl.Purpose.SERVER_AUTH, cadata=cert
        )
        if cert:
            # Disable hostname checking if and only if we have an explicit cert
            # to validate against, because the cert doesn't contain the IP addr
            # of the controller, which is what self-bootstrapped controllers
            # use. And because we pre-share and trust both the cert and
            # endpoint address anyway, it's safe to skip that check.
            # See: https://github.com/juju/python-libjuju/issues/302
            context.check_hostname = False
        return context

    async def _open(
        self, endpoint: str, cacert: str
    ) -> tuple[_WebSocket, str, str, str]:
        if self.is_debug_log_connection:
            assert self.uuid
            url = f"wss://user-{self.username}:{self.password}@{endpoint}/model/{self.uuid}/log"
        elif self.uuid:
            url = f"wss://{endpoint}/model/{self.uuid}/api"
        else:
            url = f"wss://{endpoint}/api"

        # We need to establish a server_hostname here for TLS sni if we are
        # connecting through a proxy as the Juju controller certificates will
        # not be covering the proxy
        sock = None
        server_hostname = None
        if self.proxy is not None:
            sock = self.proxy.socket()
            server_hostname = "juju-app"

        return (
            (
                await websockets.connect(
                    url,
                    ssl=self._get_ssl(cacert),
                    max_size=self.max_frame_size,
                    server_hostname=server_hostname,
                    sock=sock,
                )
            ),
            url,
            endpoint,
            cacert,
        )

    async def close(self, to_reconnect: bool = False):
        if not self._ws:
            return
        self.monitor.close_called.set()

        # Cancel all the tasks (that we started):
        tasks_need_to_be_gathered = []
        if self._pinger_task:
            tasks_need_to_be_gathered.append(self._pinger_task)
            self._pinger_task.cancel()
        if self._receiver_task:
            tasks_need_to_be_gathered.append(self._receiver_task)
            self._receiver_task.cancel()
        if self._debug_log_task:
            tasks_need_to_be_gathered.append(self._debug_log_task)
            self._debug_log_task.cancel()

        await self._ws.close()

        if not to_reconnect:
            try:
                log.debug("Gathering all tasks for connection close")
                await asyncio.gather(*tasks_need_to_be_gathered)
            except asyncio.CancelledError:
                pass
            except websockets.ConnectionClosed:
                pass

        self._pinger_task = None
        self._receiver_task = None
        self._debug_log_task = None

        if self.proxy is not None:
            self.proxy.close()

    async def _recv(self, request_id: int) -> dict[str, Any]:
        if not self.is_open:
            raise websockets.ConnectionClosedOK(rcvd=None, sent=None)
        try:
            return await self.messages.get(request_id)
        except GeneratorExit:
            return {}

    def debug_log_filter_write(self, result):
        write_or_not = True

        entity = result["tag"]
        msg_lev = result["sev"]
        mod = result["mod"]
        msg = result["msg"]

        excluded_entities = self.debug_log_params["exclude"]
        excluded_modules = self.debug_log_params["exclude_module"]
        write_or_not = (
            write_or_not
            and (mod not in excluded_modules)
            and (entity not in excluded_entities)
        )

        included_entities = self.debug_log_params["include"]
        only_these_modules = self.debug_log_params["include_module"]
        write_or_not = (
            write_or_not
            and (only_these_modules == [] or mod in only_these_modules)
            and (included_entities == [] or entity in included_entities)
        )

        log_level = self.debug_log_params["level"]

        if log_level != "" and log_level not in LEVELS:
            log.warning(
                "Debug Logger: level should be one of %s, given %s"
                % (LEVELS, log_level)
            )
        else:
            write_or_not = write_or_not and (
                log_level == "" or (LEVELS.index(msg_lev) >= LEVELS.index(log_level))
            )

        # TODO
        # lines = self.debug_log_params['lines']
        # no_tail = self.debug_log_params['no_tail']

        if write_or_not:
            ts = parse(result["ts"])

            self.debug_log_target.write(
                "%s %02d:%02d:%02d %s %s %s\n"
                % (entity, ts.hour, ts.minute, ts.second, msg_lev, mod, msg)
            )
            return 1
        else:
            return 0

    async def _debug_logger(self):
        try:
            while self.is_open:
                result = await utils.run_with_interrupt(
                    self._ws.recv(), self.monitor.close_called, log=log
                )
                if self.monitor.close_called.is_set():
                    break
                if result is not None and result != "{}\n":
                    result = json.loads(result)

                    number_of_lines_written = self.debug_log_filter_write(result)

                    self.debug_log_shown_lines += number_of_lines_written

                    if self.debug_log_shown_lines >= self.debug_log_params["limit"]:
                        asyncio.create_task(self.close(), name="Task_Close")  # noqa: RUF006
                        return

        except KeyError as e:
            log.exception("Unexpected debug line -- %s" % e)
            asyncio.create_task(self.close(), name="Task_Close")  # noqa: RUF006
            raise
        except asyncio.CancelledError:
            asyncio.create_task(self.close(), name="Task_Close")  # noqa: RUF006
            raise
        except websockets.ConnectionClosed:
            log.warning("Debug Logger: Connection closed, reconnecting")
            # the reconnect has to be done as a task because the receiver will
            # be cancelled by the reconnect and we don't want the reconnect
            # to be aborted half-way through
            asyncio.ensure_future(self.reconnect())  # noqa: RUF006
            return
        except Exception as e:
            log.exception("Error in debug logger : %s" % e)
            asyncio.create_task(self.close(), name="Task_Close")  # noqa: RUF006
            raise

    async def _receiver(self):
        try:
            while self.is_open:
                result = await utils.run_with_interrupt(
                    self._ws.recv(), self.monitor.close_called, log=log
                )
                if self.monitor.close_called.is_set():
                    break
                if result is not None:
                    result = json.loads(result)
                    await self.messages.put(result["request-id"], result)
        except asyncio.CancelledError:
            log.debug("Receiver: Cancelled")
            pass
        except websockets.ConnectionClosed as e:
            log.warning("Receiver: Connection closed, reconnecting")
            await self.messages.put_all(e)
            # the reconnect has to be done as a task because the receiver will
            # be cancelled by the reconnect and we don't want the reconnect
            # to be aborted half-way through
            asyncio.ensure_future(self.reconnect())  # noqa: RUF006
            return
        except Exception as e:
            log.exception("Error in receiver")
            # make pending listeners aware of the error
            await self.messages.put_all(e)
            raise

    async def _pinger(self):
        """A Controller can time us out if we are silent for too long. This
        is especially true in JaaS, which has a fairly strict timeout.

        To prevent timing out, we send a ping every ten seconds.

        """

        async def _do_ping():
            try:
                log.debug(f"Pinger {self._pinger_task}: pinging")
                await pinger_facade.Ping()
            except asyncio.CancelledError:
                raise

        pinger_facade = client.PingerFacade.from_connection(self)
        try:
            while True:
                await utils.run_with_interrupt(
                    _do_ping(), self.monitor.close_called, log=log
                )
                if self.monitor.close_called.is_set():
                    break
                await asyncio.sleep(10)
        except asyncio.CancelledError:
            log.debug("Pinger: Cancelled")
            pass
        except websockets.ConnectionClosed:
            # The connection has closed - we can't do anything
            # more until the connection is restarted.
            log.debug("ping failed because of closed connection")
            pass

    @overload
    async def rpc(
        self, msg: dict[str, _Json], encoder: None = None
    ) -> dict[str, _Json]: ...

    @overload
    async def rpc(
        self, msg: dict[str, _RichJson], encoder: TypeEncoder
    ) -> dict[str, _Json]: ...

    async def rpc(
        self, msg: dict[str, Any], encoder: json.JSONEncoder | None = None
    ) -> dict[str, _Json]:
        """Make an RPC to the API. The message is encoded as JSON
        using the given encoder if any.
        :param msg: Parameters for the call (will be encoded as JSON).
        :param encoder: Encoder to be used when encoding the message.
        :return: The result of the call.
        :raises JujuAPIError: When there's an error returned.
        :raises JujuError:
        """
        self.__request_id__ += 1
        msg["request-id"] = self.__request_id__
        if "params" not in msg:
            msg["params"] = {}
        if "version" not in msg:
            msg["version"] = self.facades[msg["type"]]
        outgoing = json.dumps(msg, indent=2, cls=encoder)
        log.debug(f"connection id: {id(self)} ---> {outgoing}")
        for attempt in range(3):
            if self.monitor.status == Monitor.DISCONNECTED:
                # closed cleanly; shouldn't try to reconnect
                # we're trying to use a reference to a monitor that was cleanly closed earlier
                raise websockets.ConnectionClosedOK(rcvd=None, sent=None)
            try:
                await self._ws.send(outgoing)
                break
            except websockets.ConnectionClosed:
                if attempt == 2:
                    raise
                log.warning("RPC: Connection closed, reconnecting")
                # the reconnect has to be done in a separate task because,
                # if it is triggered by the pinger, then this RPC call will
                # be cancelled when the pinger is cancelled by the reconnect,
                # and we don't want the reconnect to be aborted halfway through
                await asyncio.wait([asyncio.create_task(self.reconnect())])
                if self.monitor.status != Monitor.CONNECTED:
                    # reconnect failed; abort and shutdown
                    log.error("RPC: Automatic reconnect failed")
                    raise
        result = await self._recv(msg["request-id"])
        log.debug(f"connection id : {id(self)} <--- {result}")

        if not result:
            return result

        if "error" in result:
            # API Error Response
            raise errors.JujuAPIError(result)

        if "response" not in result:
            # This may never happen
            return result

        if "results" in result["response"]:
            # Check for errors in a result list.
            # TODO This loses the results that might have succeeded.
            # Perhaps JujuError should return all the results including
            # errors, or perhaps a keyword parameter to the rpc method
            # could be added to trigger this behaviour.
            err_results = [
                res["error"]["message"]
                for res in (result["response"]["results"] or [])
                if res.get("error", {}).get("message")
            ]
            if err_results:
                raise errors.JujuError(err_results)

        elif result["response"].get("error", {}).get("message"):
            raise errors.JujuError(result["response"]["error"]["message"])

        return result

    def _http_headers(self) -> dict[str, str]:
        """Return dictionary of http headers necessary for making an http
        connection to the endpoint of this Connection.

        :return: Dictionary of headers

        """
        if not self.usertag:
            return {}

        creds = "{}:{}".format(self.usertag, self.password or "")
        token = base64.b64encode(creds.encode())
        return {"Authorization": f"Basic {token.decode()}"}

    def https_connection(self) -> tuple[HTTPSConnection, dict[str, str], str]:
        """Return an https connection to this Connection's endpoint.

        Returns a 3-tuple containing::

            1. The :class:`HTTPSConnection` instance
            2. Dictionary of auth headers to be used with the connection
            3. The root url path (str) to be used for requests.

        """
        endpoint = self.endpoint
        # Support IPv6 by right splitting on : and removing [] around IP address for host
        host, remainder = endpoint.rsplit(":", 1)
        host = host.strip("[]")
        port = remainder
        if "/" in remainder:
            port, _ = remainder.split("/", 1)

        conn = HTTPSConnection(
            host,
            int(port),
            context=self._get_ssl(self.cacert),
        )

        path = f"/model/{self.uuid}" if self.uuid else ""
        return conn, self._http_headers(), path

    async def clone(self):
        """Return a new Connection, connected to the same websocket endpoint
        as this one.

        """
        return await Connection.connect(**self.connect_params())

    def connect_params(self):
        """Return a dict of parameters suitable for passing to
        Connection.connect that can be used to make a new connection
        to the same controller (and model if specified).
        """
        return {
            "endpoint": self.endpoint,
            "uuid": self.uuid,
            "username": self.username,
            "password": self.password,
            "cacert": self.cacert,
            "bakery_client": self.bakery_client,
            "max_frame_size": self.max_frame_size,
            "proxy": self.proxy,
        }

    async def controller(self):
        """Return a Connection to the controller at self.endpoint"""
        return await Connection.connect(
            self.endpoint,
            username=self.username,
            password=self.password,
            cacert=self.cacert,
            bakery_client=self.bakery_client,
            max_frame_size=self.max_frame_size,
        )

    async def reconnect(self):
        """Force a reconnection."""
        monitor = self.monitor
        if monitor.reconnecting.locked() or monitor.close_called.is_set():
            return
        async with monitor.reconnecting:
            await self.close(to_reconnect=True)
            connector = (
                self._connect
                if self.is_debug_log_connection
                else self._connect_with_login
            )
            res = await connector(
                [(self.endpoint, self.cacert)] if not self.endpoints else self.endpoints
            )
            if not self.is_debug_log_connection:
                self._build_facades(res.get("facades", {}))
                if not self._pinger_task:
                    log.debug("reconnect: scheduling a pinger task")
                    self._pinger_task = asyncio.create_task(
                        self._pinger(), name="Task_Pinger"
                    )

    async def _connect(self, endpoints):
        if len(endpoints) == 0:
            raise errors.JujuConnectionError("no endpoints to connect to")

        async def _try_endpoint(
            endpoint, cacert, delay
        ) -> tuple[_WebSocket, str, str, str]:
            if delay:
                await asyncio.sleep(delay)
            return await self._open(endpoint, cacert)

        # Try all endpoints in parallel, with slight increasing delay (+100ms
        # for each subsequent endpoint); the delay allows us to prefer the
        # earlier endpoints over the latter. Use first successful connection.
        tasks = [
            asyncio.ensure_future(_try_endpoint(endpoint, cacert, 0.1 * i))
            for i, (endpoint, cacert) in enumerate(endpoints)
        ]
        result: tuple[_WebSocket, str, str, str] | None = None

        for attempt in range(self._retries + 1):
            for task in asyncio.as_completed(tasks):
                try:
                    result = await task
                    break
                except ConnectionError:
                    continue  # ignore; try another endpoint
            else:
                _endpoints_str = ", ".join([endpoint for endpoint, cacert in endpoints])
                if attempt < self._retries:
                    log.debug(
                        f"Retrying connection to endpoints: {_endpoints_str}; attempt {attempt + 1} of {self._retries + 1}"
                    )
                    await asyncio.sleep((attempt + 1) * self._retry_backoff)
                    continue
                else:
                    raise errors.JujuConnectionError(
                        f"Unable to connect to any endpoint: {_endpoints_str}"
                    )
            # only executed if inner loop's else did not continue
            # (i.e., inner loop did break due to successful connection)
            break

        for task in tasks:
            task.cancel()

        assert result  # loop raises or sets the result

        self._ws = result[0]
        self.addr = result[1]
        self.endpoint = result[2]
        self.cacert = result[3]

        #  If this is a debug-log connection, and the _debug_log_task
        #  is not created yet, then go ahead and schedule it
        if self.is_debug_log_connection and not self._debug_log_task:
            self._debug_log_task = asyncio.create_task(
                self._debug_logger(), name="Task_Debug_Log"
            )

        #  If this is regular connection, and we dont have a
        #  receiver_task yet, then schedule a _receiver_task
        elif not self.is_debug_log_connection and not self._receiver_task:
            log.debug("_connect: scheduling a receiver task")
            self._receiver_task = asyncio.create_task(
                self._receiver(), name="Task_Receiver"
            )

        log.debug("Driver connected to juju %s", self.addr)
        self.monitor.close_called.clear()

    async def _connect_with_login(self, endpoints):
        """Connect to the websocket.

        If uuid is None, the connection will be to the controller. Otherwise it
        will be to the model.
        :return: The response field of login response JSON object.
        """
        success = False
        try:
            await self._connect(endpoints)
            # It's possible that we may get several discharge-required errors,
            # corresponding to different levels of authentication, so retry
            # a few times.
            for _ in range(0, 2):
                result = (await self.login())["response"]
                macaroon_json = result.get("discharge-required")
                if macaroon_json is None:
                    self.info = result
                    success = True
                    return result
                macaroon = bakery.Macaroon.from_dict(macaroon_json)
                self.bakery_client.handle_error(
                    httpbakery.Error(
                        code=httpbakery.ERR_DISCHARGE_REQUIRED,
                        message=result.get("discharge-required-error"),
                        version=macaroon.version,
                        info=httpbakery.ErrorInfo(
                            macaroon=macaroon,
                            macaroon_path=result.get("macaroon-path"),
                        ),
                    ),
                    # note: remove the port number.
                    "https://" + self.endpoint + "/",
                )
            raise errors.JujuAuthError("failed to authenticate after several attempts")
        finally:
            if not success:
                await self.close()

    async def _connect_with_redirect(self, endpoints):
        try:
            login_result = await self._connect_with_login(endpoints)
        except errors.JujuRedirectException as e:
            # Bubble up exception if the client should not follow the redirect
            if e.follow_redirect is False:
                raise
            login_result = await self._connect_with_login(e.endpoints)
        self._build_facades(login_result.get("facades", {}))
        if not self._pinger_task:
            log.debug("_connect_with_redirect: scheduling a pinger task")
            self._pinger_task = asyncio.create_task(self._pinger(), name="Task_Pinger")

    # _build_facades takes the facade list that comes from the connection with the controller,
    # validates that the client knows about them (client_facade_versions) and builds the facade list
    # (into the self._specified facades) with the max versions that both the client and the controller
    # can negotiate on
    def _build_facades(self, facades_from_connection):
        self.facades.clear()
        for facade in facades_from_connection:
            name = facade["name"]
            if name in self._specified_facades:
                client_versions = self._specified_facades[name]
            elif name in client_facade_versions:
                client_versions = client_facade_versions[name]
            elif name in known_unsupported_facades:
                continue
            else:
                log.warning(f"unexpected facade {name} received from the controller")
                continue

            controller_versions = facade["versions"]
            candidates = set(client_versions) & set(controller_versions)
            if not candidates:
                log.warning(
                    f"unknown common facade version for {name},\n"
                    f"versions known to client : {client_versions}\n"
                    f"versions known to controller : {controller_versions}"
                )
                continue
            self.facades[name] = max(candidates)

    async def login(self):
        params = {}
        # Set the client version
        params["client-version"] = CLIENT_VERSION
        params["auth-tag"] = self.usertag
        if self.password:
            params["credentials"] = self.password
        else:
            macaroons = _macaroons_for_domain(self.bakery_client.cookies, self.endpoint)
            params["macaroons"] = [
                [bakery.macaroon_to_dict(m) for m in ms] for ms in macaroons
            ]

        try:
            return await self.rpc({
                "type": "Admin",
                "request": "Login",
                "version": 3,
                "params": params,
            })
        except errors.JujuAPIError as e:
            if e.error_code != "redirection required":
                raise
            log.info("Controller requested redirect")
            # Check if the redirect error provides a payload with embedded
            # redirection info (juju 2.6+ controller). In this case, return a
            # redirect exception which the library should not automatically
            # follow but rather bubble up to the user. This matches the
            # behaviour of juju cli whereas for JAAS-like redirects we will
            # need to make an extra RPC call to get the redirect info.
            if e.error_info is not None:
                raise errors.JujuRedirectException(e.error_info, False) from e

            # Fetch additional redirection information now so that
            # we can safely close the connection after login
            # fails.
            redirect_info = (
                await self.rpc({
                    "type": "Admin",
                    "request": "RedirectInfo",
                    "version": 3,
                })
            )["response"]
            raise errors.JujuRedirectException(redirect_info, True) from e


def _macaroons_for_domain(cookies, domain):
    """Return any macaroons from the given cookie jar that
    apply to the given domain name.
    """
    req = urllib.request.Request("https://" + domain + "/")
    cookies.add_cookie_header(req)
    return httpbakery.extract_macaroons(req)



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class ActionFacade(Type):
    name = "Action"
    version = 7

    @ReturnMapping(ActionResults)
    async def Actions(self, entities=None):
        """Actions takes a list of ActionTags, and returns the full Action for
        each ID.

        entities : typing.Sequence[~Entity]
        Returns -> ActionResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Action", request="Actions", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ApplicationsCharmActionsResults)
    async def ApplicationsCharmsActions(self, entities=None):
        """ApplicationsCharmsActions returns a slice of charm Actions for a slice of
        services.

        entities : typing.Sequence[~Entity]
        Returns -> ApplicationsCharmActionsResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Action",
            request="ApplicationsCharmsActions",
            version=7,
            params=_params,
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ActionResults)
    async def Cancel(self, entities=None):
        """Cancel attempts to cancel enqueued Actions from running.

        entities : typing.Sequence[~Entity]
        Returns -> ActionResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Action", request="Cancel", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(EnqueuedActions)
    async def EnqueueOperation(self, actions=None):
        """EnqueueOperation takes a list of Actions and queues them up to be executed as
        an operation, each action running as a task on the designated ActionReceiver.
        We return the ID of the overall operation and each individual task.

        actions : typing.Sequence[~Action]
        Returns -> EnqueuedActions
        """
        if actions is not None and not isinstance(actions, (bytes, str, list)):
            raise Exception(
                f"Expected actions to be a Sequence, received: {type(actions)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Action", request="EnqueueOperation", version=7, params=_params)
        _params["actions"] = actions
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(OperationResults)
    async def ListOperations(
        self,
        actions=None,
        applications=None,
        limit=None,
        machines=None,
        offset=None,
        status=None,
        units=None,
    ):
        """ListOperations fetches the called actions for specified apps/units.

        actions : typing.Sequence[str]
        applications : typing.Sequence[str]
        limit : int
        machines : typing.Sequence[str]
        offset : int
        status : typing.Sequence[str]
        units : typing.Sequence[str]
        Returns -> OperationResults
        """
        if actions is not None and not isinstance(actions, (bytes, str, list)):
            raise Exception(
                f"Expected actions to be a Sequence, received: {type(actions)}"
            )

        if applications is not None and not isinstance(
            applications, (bytes, str, list)
        ):
            raise Exception(
                f"Expected applications to be a Sequence, received: {type(applications)}"
            )

        if limit is not None and not isinstance(limit, int):
            raise Exception(f"Expected limit to be a int, received: {type(limit)}")

        if machines is not None and not isinstance(machines, (bytes, str, list)):
            raise Exception(
                f"Expected machines to be a Sequence, received: {type(machines)}"
            )

        if offset is not None and not isinstance(offset, int):
            raise Exception(f"Expected offset to be a int, received: {type(offset)}")

        if status is not None and not isinstance(status, (bytes, str, list)):
            raise Exception(
                f"Expected status to be a Sequence, received: {type(status)}"
            )

        if units is not None and not isinstance(units, (bytes, str, list)):
            raise Exception(f"Expected units to be a Sequence, received: {type(units)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Action", request="ListOperations", version=7, params=_params)
        _params["actions"] = actions
        _params["applications"] = applications
        _params["limit"] = limit
        _params["machines"] = machines
        _params["offset"] = offset
        _params["status"] = status
        _params["units"] = units
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(OperationResults)
    async def Operations(self, entities=None):
        """Operations fetches the specified operation ids.

        entities : typing.Sequence[~Entity]
        Returns -> OperationResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Action", request="Operations", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(EnqueuedActions)
    async def Run(
        self,
        applications=None,
        commands=None,
        execution_group=None,
        machines=None,
        parallel=None,
        timeout=None,
        units=None,
        workload_context=None,
    ):
        """Run the commands specified on the machines identified through the
        list of machines, units and services.

        applications : typing.Sequence[str]
        commands : str
        execution_group : str
        machines : typing.Sequence[str]
        parallel : bool
        timeout : int
        units : typing.Sequence[str]
        workload_context : bool
        Returns -> EnqueuedActions
        """
        if applications is not None and not isinstance(
            applications, (bytes, str, list)
        ):
            raise Exception(
                f"Expected applications to be a Sequence, received: {type(applications)}"
            )

        if commands is not None and not isinstance(commands, (bytes, str)):
            raise Exception(
                f"Expected commands to be a str, received: {type(commands)}"
            )

        if execution_group is not None and not isinstance(
            execution_group, (bytes, str)
        ):
            raise Exception(
                f"Expected execution_group to be a str, received: {type(execution_group)}"
            )

        if machines is not None and not isinstance(machines, (bytes, str, list)):
            raise Exception(
                f"Expected machines to be a Sequence, received: {type(machines)}"
            )

        if parallel is not None and not isinstance(parallel, bool):
            raise Exception(
                f"Expected parallel to be a bool, received: {type(parallel)}"
            )

        if timeout is not None and not isinstance(timeout, int):
            raise Exception(f"Expected timeout to be a int, received: {type(timeout)}")

        if units is not None and not isinstance(units, (bytes, str, list)):
            raise Exception(f"Expected units to be a Sequence, received: {type(units)}")

        if workload_context is not None and not isinstance(workload_context, bool):
            raise Exception(
                f"Expected workload_context to be a bool, received: {type(workload_context)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Action", request="Run", version=7, params=_params)
        _params["applications"] = applications
        _params["commands"] = commands
        _params["execution-group"] = execution_group
        _params["machines"] = machines
        _params["parallel"] = parallel
        _params["timeout"] = timeout
        _params["units"] = units
        _params["workload-context"] = workload_context
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(EnqueuedActions)
    async def RunOnAllMachines(
        self,
        applications=None,
        commands=None,
        execution_group=None,
        machines=None,
        parallel=None,
        timeout=None,
        units=None,
        workload_context=None,
    ):
        """RunOnAllMachines attempts to run the specified command on all the machines.

        applications : typing.Sequence[str]
        commands : str
        execution_group : str
        machines : typing.Sequence[str]
        parallel : bool
        timeout : int
        units : typing.Sequence[str]
        workload_context : bool
        Returns -> EnqueuedActions
        """
        if applications is not None and not isinstance(
            applications, (bytes, str, list)
        ):
            raise Exception(
                f"Expected applications to be a Sequence, received: {type(applications)}"
            )

        if commands is not None and not isinstance(commands, (bytes, str)):
            raise Exception(
                f"Expected commands to be a str, received: {type(commands)}"
            )

        if execution_group is not None and not isinstance(
            execution_group, (bytes, str)
        ):
            raise Exception(
                f"Expected execution_group to be a str, received: {type(execution_group)}"
            )

        if machines is not None and not isinstance(machines, (bytes, str, list)):
            raise Exception(
                f"Expected machines to be a Sequence, received: {type(machines)}"
            )

        if parallel is not None and not isinstance(parallel, bool):
            raise Exception(
                f"Expected parallel to be a bool, received: {type(parallel)}"
            )

        if timeout is not None and not isinstance(timeout, int):
            raise Exception(f"Expected timeout to be a int, received: {type(timeout)}")

        if units is not None and not isinstance(units, (bytes, str, list)):
            raise Exception(f"Expected units to be a Sequence, received: {type(units)}")

        if workload_context is not None and not isinstance(workload_context, bool):
            raise Exception(
                f"Expected workload_context to be a bool, received: {type(workload_context)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Action", request="RunOnAllMachines", version=7, params=_params)
        _params["applications"] = applications
        _params["commands"] = commands
        _params["execution-group"] = execution_group
        _params["machines"] = machines
        _params["parallel"] = parallel
        _params["timeout"] = timeout
        _params["units"] = units
        _params["workload-context"] = workload_context
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringsWatchResults)
    async def WatchActionsProgress(self, entities=None):
        """WatchActionsProgress creates a watcher that reports on action log messages.

        entities : typing.Sequence[~Entity]
        Returns -> StringsWatchResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Action", request="WatchActionsProgress", version=7, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply


class CharmsFacade(Type):
    name = "Charms"
    version = 7

    @ReturnMapping(CharmOriginResult)
    async def AddCharm(self, charm_origin=None, force=None, url=None):
        """AddCharm adds the given charm URL (which must include revision) to the
        environment, if it does not exist yet. Local charms are not supported,
        only charm store and charm hub URLs. See also AddLocalCharm().

        charm_origin : CharmOrigin
        force : bool
        url : str
        Returns -> CharmOriginResult
        """
        if charm_origin is not None and not isinstance(
            charm_origin, (dict, CharmOrigin)
        ):
            raise Exception(
                f"Expected charm_origin to be a CharmOrigin, received: {type(charm_origin)}"
            )

        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if url is not None and not isinstance(url, (bytes, str)):
            raise Exception(f"Expected url to be a str, received: {type(url)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="AddCharm", version=7, params=_params)
        _params["charm-origin"] = charm_origin
        _params["force"] = force
        _params["url"] = url
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(Charm)
    async def CharmInfo(self, url=None):
        """CharmInfo returns information about the requested charm.

        url : str
        Returns -> Charm
        """
        if url is not None and not isinstance(url, (bytes, str)):
            raise Exception(f"Expected url to be a str, received: {type(url)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="CharmInfo", version=7, params=_params)
        _params["url"] = url
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def CheckCharmPlacement(self, placements=None):
        """CheckCharmPlacement checks if a charm is allowed to be placed with in a
        given application.

        placements : typing.Sequence[~ApplicationCharmPlacement]
        Returns -> ErrorResults
        """
        if placements is not None and not isinstance(placements, (bytes, str, list)):
            raise Exception(
                f"Expected placements to be a Sequence, received: {type(placements)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Charms", request="CheckCharmPlacement", version=7, params=_params
        )
        _params["placements"] = placements
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(DownloadInfoResults)
    async def GetDownloadInfos(self, entities=None):
        """GetDownloadInfos attempts to get the bundle corresponding to the charm url
        and origin.

        entities : typing.Sequence[~CharmURLAndOrigin]
        Returns -> DownloadInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="GetDownloadInfos", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(IsMeteredResult)
    async def IsMetered(self, url=None):
        """IsMetered returns whether or not the charm is metered.
        TODO (cderici) only used for metered charms in cmd MeteredDeployAPI,
        kept for client compatibility, remove in juju 4.0

        url : str
        Returns -> IsMeteredResult
        """
        if url is not None and not isinstance(url, (bytes, str)):
            raise Exception(f"Expected url to be a str, received: {type(url)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="IsMetered", version=7, params=_params)
        _params["url"] = url
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CharmsListResult)
    async def List(self, names=None):
        """List returns a list of charm URLs currently in the state.
        If supplied parameter contains any names, the result will
        be filtered to return only the charms with supplied names.

        names : typing.Sequence[str]
        Returns -> CharmsListResult
        """
        if names is not None and not isinstance(names, (bytes, str, list)):
            raise Exception(f"Expected names to be a Sequence, received: {type(names)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="List", version=7, params=_params)
        _params["names"] = names
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CharmResourcesResults)
    async def ListCharmResources(self, entities=None):
        """ListCharmResources returns a series of resources for a given charm.

        entities : typing.Sequence[~CharmURLAndOrigin]
        Returns -> CharmResourcesResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Charms", request="ListCharmResources", version=7, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ResolveCharmWithChannelResults)
    async def ResolveCharms(self, macaroon=None, resolve=None):
        """ResolveCharms resolves the given charm URLs with an optionally specified
        preferred channel.  Channel provided via CharmOrigin.

        macaroon : Macaroon
        resolve : typing.Sequence[~ResolveCharmWithChannel]
        Returns -> ResolveCharmWithChannelResults
        """
        if macaroon is not None and not isinstance(macaroon, (dict, Macaroon)):
            raise Exception(
                f"Expected macaroon to be a Macaroon, received: {type(macaroon)}"
            )

        if resolve is not None and not isinstance(resolve, (bytes, str, list)):
            raise Exception(
                f"Expected resolve to be a Sequence, received: {type(resolve)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="ResolveCharms", version=7, params=_params)
        _params["macaroon"] = macaroon
        _params["resolve"] = resolve
        reply = await self.rpc(msg)
        return reply


class ClientFacade(Type):
    name = "Client"
    version = 7

    @ReturnMapping(FindToolsResult)
    async def FindTools(
        self, agentstream=None, arch=None, major=None, number=None, os_type=None
    ):
        """FindTools returns a List containing all tools matching the given parameters.
        TODO(juju 3.1) - remove, used by 2.9 client only

        agentstream : str
        arch : str
        major : int
        number : Number
        os_type : str
        Returns -> FindToolsResult
        """
        if agentstream is not None and not isinstance(agentstream, (bytes, str)):
            raise Exception(
                f"Expected agentstream to be a str, received: {type(agentstream)}"
            )

        if arch is not None and not isinstance(arch, (bytes, str)):
            raise Exception(f"Expected arch to be a str, received: {type(arch)}")

        if major is not None and not isinstance(major, int):
            raise Exception(f"Expected major to be a int, received: {type(major)}")

        if number is not None and not isinstance(number, (dict, Number)):
            raise Exception(f"Expected number to be a Number, received: {type(number)}")

        if os_type is not None and not isinstance(os_type, (bytes, str)):
            raise Exception(f"Expected os_type to be a str, received: {type(os_type)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="FindTools", version=7, params=_params)
        _params["agentstream"] = agentstream
        _params["arch"] = arch
        _params["major"] = major
        _params["number"] = number
        _params["os-type"] = os_type
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(FullStatus)
    async def FullStatus(self, include_storage=None, patterns=None):
        """FullStatus gives the information needed for juju status over the api

        include_storage : bool
        patterns : typing.Sequence[str]
        Returns -> FullStatus
        """
        if include_storage is not None and not isinstance(include_storage, bool):
            raise Exception(
                f"Expected include_storage to be a bool, received: {type(include_storage)}"
            )

        if patterns is not None and not isinstance(patterns, (bytes, str, list)):
            raise Exception(
                f"Expected patterns to be a Sequence, received: {type(patterns)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="FullStatus", version=7, params=_params)
        _params["include-storage"] = include_storage
        _params["patterns"] = patterns
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StatusHistoryResults)
    async def StatusHistory(self, requests=None):
        """StatusHistory returns a slice of past statuses for several entities.

        requests : typing.Sequence[~StatusHistoryRequest]
        Returns -> StatusHistoryResults
        """
        if requests is not None and not isinstance(requests, (bytes, str, list)):
            raise Exception(
                f"Expected requests to be a Sequence, received: {type(requests)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="StatusHistory", version=7, params=_params)
        _params["requests"] = requests
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(AllWatcherId)
    async def WatchAll(self):
        """WatchAll initiates a watcher for entities in the connected model.

        Returns -> AllWatcherId
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="WatchAll", version=7, params=_params)

        reply = await self.rpc(msg)
        return reply


class CloudFacade(Type):
    name = "Cloud"
    version = 7

    @ReturnMapping(None)
    async def AddCloud(self, cloud=None, force=None, name=None):
        """AddCloud adds a new cloud, different from the one managed by the controller.

        cloud : Cloud
        force : bool
        name : str
        Returns -> None
        """
        if cloud is not None and not isinstance(cloud, (dict, Cloud)):
            raise Exception(f"Expected cloud to be a Cloud, received: {type(cloud)}")

        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if name is not None and not isinstance(name, (bytes, str)):
            raise Exception(f"Expected name to be a str, received: {type(name)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="AddCloud", version=7, params=_params)
        _params["cloud"] = cloud
        _params["force"] = force
        _params["name"] = name
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def AddCredentials(self, credentials=None):
        """AddCredentials adds new credentials.
        In contrast to UpdateCredentials() below, the new credentials can be
        for a cloud that the controller does not manage (this is required
        for CAAS models)

        credentials : typing.Sequence[~TaggedCredential]
        Returns -> ErrorResults
        """
        if credentials is not None and not isinstance(credentials, (bytes, str, list)):
            raise Exception(
                f"Expected credentials to be a Sequence, received: {type(credentials)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="AddCredentials", version=7, params=_params)
        _params["credentials"] = credentials
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(UpdateCredentialResults)
    async def CheckCredentialsModels(self, credentials=None):
        """CheckCredentialsModels validates supplied cloud credentials' content against
        models that currently use these credentials.
        If there are any models that are using a credential and these models or their
        cloud instances are not going to be accessible with corresponding credential,
        there will be detailed validation errors per model.
        There's no Juju API client which uses this, but JAAS does,

        credentials : typing.Sequence[~TaggedCredential]
        Returns -> UpdateCredentialResults
        """
        if credentials is not None and not isinstance(credentials, (bytes, str, list)):
            raise Exception(
                f"Expected credentials to be a Sequence, received: {type(credentials)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Cloud", request="CheckCredentialsModels", version=7, params=_params
        )
        _params["credentials"] = credentials
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CloudResults)
    async def Cloud(self, entities=None):
        """Cloud returns the cloud definitions for the specified clouds.

        entities : typing.Sequence[~Entity]
        Returns -> CloudResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="Cloud", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CloudInfoResults)
    async def CloudInfo(self, entities=None):
        """CloudInfo returns information about the specified clouds.

        entities : typing.Sequence[~Entity]
        Returns -> CloudInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="CloudInfo", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CloudsResult)
    async def Clouds(self):
        """Clouds returns the definitions of all clouds supported by the controller
        that the logged in user can see.

        Returns -> CloudsResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="Clouds", version=7, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CloudCredentialResults)
    async def Credential(self, entities=None):
        """Credential returns the specified cloud credential for each tag, minus secrets.

        entities : typing.Sequence[~Entity]
        Returns -> CloudCredentialResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="Credential", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CredentialContentResults)
    async def CredentialContents(self, credentials=None, include_secrets=None):
        """CredentialContents returns the specified cloud credentials,
        including the secrets if requested.
        If no specific credential name/cloud was passed in, all credentials for this user
        are returned.
        Only credential owner can see its contents as well as what models use it.
        Controller admin has no special superpowers here and is treated the same as all other users.

        credentials : typing.Sequence[~CloudCredentialArg]
        include_secrets : bool
        Returns -> CredentialContentResults
        """
        if credentials is not None and not isinstance(credentials, (bytes, str, list)):
            raise Exception(
                f"Expected credentials to be a Sequence, received: {type(credentials)}"
            )

        if include_secrets is not None and not isinstance(include_secrets, bool):
            raise Exception(
                f"Expected include_secrets to be a bool, received: {type(include_secrets)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Cloud", request="CredentialContents", version=7, params=_params
        )
        _params["credentials"] = credentials
        _params["include-secrets"] = include_secrets
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(InstanceTypesResults)
    async def InstanceTypes(self, constraints=None):
        """InstanceTypes returns instance type information for the cloud and region
        in which the current model is deployed.

        constraints : typing.Sequence[~CloudInstanceTypesConstraint]
        Returns -> InstanceTypesResults
        """
        if constraints is not None and not isinstance(constraints, (bytes, str, list)):
            raise Exception(
                f"Expected constraints to be a Sequence, received: {type(constraints)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="InstanceTypes", version=7, params=_params)
        _params["constraints"] = constraints
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ListCloudInfoResults)
    async def ListCloudInfo(self, all_=None, user_tag=None):
        """ListCloudInfo returns clouds that the specified user has access to.
        Controller admins (superuser) can list clouds for any user.
        Other users can only ask about their own clouds.

        all_ : bool
        user_tag : str
        Returns -> ListCloudInfoResults
        """
        if all_ is not None and not isinstance(all_, bool):
            raise Exception(f"Expected all_ to be a bool, received: {type(all_)}")

        if user_tag is not None and not isinstance(user_tag, (bytes, str)):
            raise Exception(
                f"Expected user_tag to be a str, received: {type(user_tag)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="ListCloudInfo", version=7, params=_params)
        _params["all"] = all_
        _params["user-tag"] = user_tag
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def ModifyCloudAccess(self, changes=None):
        """ModifyCloudAccess changes the model access granted to users.

        changes : typing.Sequence[~ModifyCloudAccess]
        Returns -> ErrorResults
        """
        if changes is not None and not isinstance(changes, (bytes, str, list)):
            raise Exception(
                f"Expected changes to be a Sequence, received: {type(changes)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="ModifyCloudAccess", version=7, params=_params)
        _params["changes"] = changes
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def RemoveClouds(self, entities=None):
        """RemoveClouds removes the specified clouds from the controller.
        If a cloud is in use (has models deployed to it), the removal will fail.

        entities : typing.Sequence[~Entity]
        Returns -> ErrorResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="RemoveClouds", version=7, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def RevokeCredentialsCheckModels(self, credentials=None):
        """RevokeCredentialsCheckModels revokes a set of cloud credentials.
        If the credentials are used by any of the models, the credential deletion will be aborted.
        If credential-in-use needs to be revoked nonetheless, this method allows the use of force.

        credentials : typing.Sequence[~RevokeCredentialArg]
        Returns -> ErrorResults
        """
        if credentials is not None and not isinstance(credentials, (bytes, str, list)):
            raise Exception(
                f"Expected credentials to be a Sequence, received: {type(credentials)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Cloud",
            request="RevokeCredentialsCheckModels",
            version=7,
            params=_params,
        )
        _params["credentials"] = credentials
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def UpdateCloud(self, clouds=None):
        """UpdateCloud updates an existing cloud that the controller knows about.

        clouds : typing.Sequence[~AddCloudArgs]
        Returns -> ErrorResults
        """
        if clouds is not None and not isinstance(clouds, (bytes, str, list)):
            raise Exception(
                f"Expected clouds to be a Sequence, received: {type(clouds)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="UpdateCloud", version=7, params=_params)
        _params["clouds"] = clouds
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(UpdateCredentialResults)
    async def UpdateCredentialsCheckModels(self, credentials=None, force=None):
        """UpdateCredentialsCheckModels updates a set of cloud credentials' content.
        If there are any models that are using a credential and these models
        are not going to be visible with updated credential content,
        there will be detailed validation errors per model.  Such model errors are returned
        separately and do not contribute to the overall method error status.
        Controller admins can 'force' an update of the credential
        regardless of whether it is deemed valid or not.

        credentials : typing.Sequence[~TaggedCredential]
        force : bool
        Returns -> UpdateCredentialResults
        """
        if credentials is not None and not isinstance(credentials, (bytes, str, list)):
            raise Exception(
                f"Expected credentials to be a Sequence, received: {type(credentials)}"
            )

        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Cloud",
            request="UpdateCredentialsCheckModels",
            version=7,
            params=_params,
        )
        _params["credentials"] = credentials
        _params["force"] = force
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringsResults)
    async def UserCredentials(self, user_clouds=None):
        """UserCredentials returns the cloud credentials for a set of users.

        user_clouds : typing.Sequence[~UserCloud]
        Returns -> StringsResults
        """
        if user_clouds is not None and not isinstance(user_clouds, (bytes, str, list)):
            raise Exception(
                f"Expected user_clouds to be a Sequence, received: {type(user_clouds)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Cloud", request="UserCredentials", version=7, params=_params)
        _params["user-clouds"] = user_clouds
        reply = await self.rpc(msg)
        return reply



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
from __future__ import annotations

from . import model
from .client import client, overrides


def get_entity_delta(d: overrides.Delta):
    return _delta_types[d.entity](d.deltas)


def get_entity_class(entity_type):
    return _delta_types[entity_type].get_entity_class()


class EntityDelta(client.Delta):
    data: dict[str, str]

    def get_id(self) -> str:
        return self.data["id"]

    @classmethod
    def get_entity_class(cls) -> type[model.ModelEntity]:
        raise NotImplementedError()


class ActionDelta(EntityDelta):
    @classmethod
    def get_entity_class(cls):
        from .action import Action

        return Action


class ApplicationDelta(EntityDelta):
    def get_id(self):
        return self.data["name"]

    @classmethod
    def get_entity_class(cls):
        from .application import Application

        return Application


class AnnotationDelta(EntityDelta):
    def get_id(self):
        return self.data["tag"]

    @classmethod
    def get_entity_class(cls):
        from .annotation import Annotation

        return Annotation


class ModelDelta(EntityDelta):
    def get_id(self):
        return self.data["model-uuid"]

    @classmethod
    def get_entity_class(cls):
        from .model import ModelInfo

        return ModelInfo


class MachineDelta(EntityDelta):
    @classmethod
    def get_entity_class(cls):
        from .machine import Machine

        return Machine


class UnitDelta(EntityDelta):
    def get_id(self):
        return self.data["name"]

    @classmethod
    def get_entity_class(cls):
        from .unit import Unit

        return Unit


class RelationDelta(EntityDelta):
    @classmethod
    def get_entity_class(cls):
        from .relation import Relation

        return Relation


class RemoteApplicationDelta(EntityDelta):
    def get_id(self):
        return self.data["name"]

    @classmethod
    def get_entity_class(cls):
        from .remoteapplication import RemoteApplication

        return RemoteApplication


class CharmDelta(EntityDelta):
    def get_id(self):
        return self.data["charm-url"]

    @classmethod
    def get_entity_class(cls):
        from .charm import Charm

        return Charm


class ApplicationOfferDelta(EntityDelta):
    def get_id(self):
        return self.data["application-name"]

    @classmethod
    def get_entity_class(cls):
        from .remoteapplication import ApplicationOffer

        return ApplicationOffer


_delta_types = {
    "action": ActionDelta,
    "annotation": AnnotationDelta,
    "application": ApplicationDelta,
    "applicationOffer": ApplicationOfferDelta,
    "charm": CharmDelta,
    "machine": MachineDelta,
    "model": ModelDelta,
    "relation": RelationDelta,
    "remoteApplication": RemoteApplicationDelta,
    "unit": UnitDelta,
}



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio

from juju.model import Model


async def main():
    m = Model()
    await m.connect()

    secrets = await m.list_secrets()
    print(secrets)
    await m.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:
1. Connects to the current model
2. Deploy a bundle from charmhub and waits until it reports itself active
3. Destroys the unit and application
"""

import asyncio

from juju.model import Model


async def main():
    model = Model()
    print("Connecting to model")
    # connect to current model with current user, per Juju CLI
    await model.connect()

    try:
        print("Deploying lma-light")
        applications = await model.deploy("ch:lma-light", channel="edge", trust=True)

        print("Waiting for active")
        await model.wait_for_idle(status="active")

        print("Removing lma-light")
        for application in applications:
            await application.remove()
    finally:
        print("Disconnecting from model")
        await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class CredentialManagerFacade(Type):
    name = "CredentialManager"
    version = 1

    @ReturnMapping(ErrorResult)
    async def InvalidateModelCredential(self, reason=None):
        """InvalidateModelCredential marks the cloud credential for this model as invalid.

        reason : str
        Returns -> ErrorResult
        """
        if reason is not None and not isinstance(reason, (bytes, str)):
            raise Exception(f"Expected reason to be a str, received: {type(reason)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="CredentialManager",
            request="InvalidateModelCredential",
            version=1,
            params=_params,
        )
        _params["reason"] = reason
        reply = await self.rpc(msg)
        return reply


class FirewallRulesFacade(Type):
    name = "FirewallRules"
    version = 1

    @ReturnMapping(ListFirewallRulesResults)
    async def ListFirewallRules(self):
        """ListFirewallRules returns all the firewall rules.

        Returns -> ListFirewallRulesResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="FirewallRules", request="ListFirewallRules", version=1, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def SetFirewallRules(self, args=None):
        """SetFirewallRules creates or updates the specified firewall rules.

        args : typing.Sequence[~FirewallRule]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="FirewallRules", request="SetFirewallRules", version=1, params=_params
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply


class ImageMetadataManagerFacade(Type):
    name = "ImageMetadataManager"
    version = 1

    @ReturnMapping(ErrorResults)
    async def Delete(self, image_ids=None):
        """Delete deletes cloud image metadata for given image ids.
        It supports bulk calls.

        image_ids : typing.Sequence[str]
        Returns -> ErrorResults
        """
        if image_ids is not None and not isinstance(image_ids, (bytes, str, list)):
            raise Exception(
                f"Expected image_ids to be a Sequence, received: {type(image_ids)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ImageMetadataManager", request="Delete", version=1, params=_params
        )
        _params["image-ids"] = image_ids
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ListCloudImageMetadataResult)
    async def List(
        self,
        arches=None,
        region=None,
        root_storage_type=None,
        stream=None,
        versions=None,
        virt_type=None,
    ):
        """List returns all found cloud image metadata that satisfy
        given filter.
        Returned list contains metadata ordered by priority.

        arches : typing.Sequence[str]
        region : str
        root_storage_type : str
        stream : str
        versions : typing.Sequence[str]
        virt_type : str
        Returns -> ListCloudImageMetadataResult
        """
        if arches is not None and not isinstance(arches, (bytes, str, list)):
            raise Exception(
                f"Expected arches to be a Sequence, received: {type(arches)}"
            )

        if region is not None and not isinstance(region, (bytes, str)):
            raise Exception(f"Expected region to be a str, received: {type(region)}")

        if root_storage_type is not None and not isinstance(
            root_storage_type, (bytes, str)
        ):
            raise Exception(
                f"Expected root_storage_type to be a str, received: {type(root_storage_type)}"
            )

        if stream is not None and not isinstance(stream, (bytes, str)):
            raise Exception(f"Expected stream to be a str, received: {type(stream)}")

        if versions is not None and not isinstance(versions, (bytes, str, list)):
            raise Exception(
                f"Expected versions to be a Sequence, received: {type(versions)}"
            )

        if virt_type is not None and not isinstance(virt_type, (bytes, str)):
            raise Exception(
                f"Expected virt_type to be a str, received: {type(virt_type)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ImageMetadataManager", request="List", version=1, params=_params
        )
        _params["arches"] = arches
        _params["region"] = region
        _params["root-storage-type"] = root_storage_type
        _params["stream"] = stream
        _params["versions"] = versions
        _params["virt-type"] = virt_type
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def Save(self, metadata=None):
        """Save stores given cloud image metadata.
        It supports bulk calls.

        metadata : typing.Sequence[~CloudImageMetadataList]
        Returns -> ErrorResults
        """
        if metadata is not None and not isinstance(metadata, (bytes, str, list)):
            raise Exception(
                f"Expected metadata to be a Sequence, received: {type(metadata)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ImageMetadataManager", request="Save", version=1, params=_params
        )
        _params["metadata"] = metadata
        reply = await self.rpc(msg)
        return reply


class KeyManagerFacade(Type):
    name = "KeyManager"
    version = 1

    @ReturnMapping(ErrorResults)
    async def AddKeys(self, ssh_keys=None, user=None):
        """AddKeys adds new authorised ssh keys for the specified user.

        ssh_keys : typing.Sequence[str]
        user : str
        Returns -> ErrorResults
        """
        if ssh_keys is not None and not isinstance(ssh_keys, (bytes, str, list)):
            raise Exception(
                f"Expected ssh_keys to be a Sequence, received: {type(ssh_keys)}"
            )

        if user is not None and not isinstance(user, (bytes, str)):
            raise Exception(f"Expected user to be a str, received: {type(user)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="KeyManager", request="AddKeys", version=1, params=_params)
        _params["ssh-keys"] = ssh_keys
        _params["user"] = user
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def DeleteKeys(self, ssh_keys=None, user=None):
        """DeleteKeys deletes the authorised ssh keys for the specified user.

        ssh_keys : typing.Sequence[str]
        user : str
        Returns -> ErrorResults
        """
        if ssh_keys is not None and not isinstance(ssh_keys, (bytes, str, list)):
            raise Exception(
                f"Expected ssh_keys to be a Sequence, received: {type(ssh_keys)}"
            )

        if user is not None and not isinstance(user, (bytes, str)):
            raise Exception(f"Expected user to be a str, received: {type(user)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="KeyManager", request="DeleteKeys", version=1, params=_params)
        _params["ssh-keys"] = ssh_keys
        _params["user"] = user
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def ImportKeys(self, ssh_keys=None, user=None):
        """ImportKeys imports new authorised ssh keys from the specified key ids for the specified user.

        ssh_keys : typing.Sequence[str]
        user : str
        Returns -> ErrorResults
        """
        if ssh_keys is not None and not isinstance(ssh_keys, (bytes, str, list)):
            raise Exception(
                f"Expected ssh_keys to be a Sequence, received: {type(ssh_keys)}"
            )

        if user is not None and not isinstance(user, (bytes, str)):
            raise Exception(f"Expected user to be a str, received: {type(user)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="KeyManager", request="ImportKeys", version=1, params=_params)
        _params["ssh-keys"] = ssh_keys
        _params["user"] = user
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringsResults)
    async def ListKeys(self, entities=None, mode=None):
        """ListKeys returns the authorised ssh keys for the specified users.

        entities : Entities
        mode : bool
        Returns -> StringsResults
        """
        if entities is not None and not isinstance(entities, (dict, Entities)):
            raise Exception(
                f"Expected entities to be a Entities, received: {type(entities)}"
            )

        if mode is not None and not isinstance(mode, bool):
            raise Exception(f"Expected mode to be a bool, received: {type(mode)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="KeyManager", request="ListKeys", version=1, params=_params)
        _params["entities"] = entities
        _params["mode"] = mode
        reply = await self.rpc(msg)
        return reply


class ModelUpgraderFacade(Type):
    name = "ModelUpgrader"
    version = 1

    @ReturnMapping(None)
    async def AbortModelUpgrade(self, model_tag=None):
        """AbortModelUpgrade aborts and archives the model upgrade
        synchronisation record, if any.

        model_tag : str
        Returns -> None
        """
        if model_tag is not None and not isinstance(model_tag, (bytes, str)):
            raise Exception(
                f"Expected model_tag to be a str, received: {type(model_tag)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelUpgrader", request="AbortModelUpgrade", version=1, params=_params
        )
        _params["model-tag"] = model_tag
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(UpgradeModelResult)
    async def UpgradeModel(
        self,
        agent_stream=None,
        dry_run=None,
        ignore_agent_versions=None,
        model_tag=None,
        target_version=None,
    ):
        """UpgradeModel upgrades a model.

        agent_stream : str
        dry_run : bool
        ignore_agent_versions : bool
        model_tag : str
        target_version : Number
        Returns -> UpgradeModelResult
        """
        if agent_stream is not None and not isinstance(agent_stream, (bytes, str)):
            raise Exception(
                f"Expected agent_stream to be a str, received: {type(agent_stream)}"
            )

        if dry_run is not None and not isinstance(dry_run, bool):
            raise Exception(f"Expected dry_run to be a bool, received: {type(dry_run)}")

        if ignore_agent_versions is not None and not isinstance(
            ignore_agent_versions, bool
        ):
            raise Exception(
                f"Expected ignore_agent_versions to be a bool, received: {type(ignore_agent_versions)}"
            )

        if model_tag is not None and not isinstance(model_tag, (bytes, str)):
            raise Exception(
                f"Expected model_tag to be a str, received: {type(model_tag)}"
            )

        if target_version is not None and not isinstance(
            target_version, (dict, Number)
        ):
            raise Exception(
                f"Expected target_version to be a Number, received: {type(target_version)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelUpgrader", request="UpgradeModel", version=1, params=_params
        )
        _params["agent-stream"] = agent_stream
        _params["dry-run"] = dry_run
        _params["ignore-agent-versions"] = ignore_agent_versions
        _params["model-tag"] = model_tag
        _params["target-version"] = target_version
        reply = await self.rpc(msg)
        return reply


class PayloadsFacade(Type):
    name = "Payloads"
    version = 1

    @ReturnMapping(PayloadListResults)
    async def List(self, patterns=None):
        """List builds the list of payloads being tracked for
        the given unit and IDs. If no IDs are provided then all tracked
        payloads for the unit are returned.

        patterns : typing.Sequence[str]
        Returns -> PayloadListResults
        """
        if patterns is not None and not isinstance(patterns, (bytes, str, list)):
            raise Exception(
                f"Expected patterns to be a Sequence, received: {type(patterns)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Payloads", request="List", version=1, params=_params)
        _params["patterns"] = patterns
        reply = await self.rpc(msg)
        return reply


class PingerFacade(Type):
    name = "Pinger"
    version = 1

    @ReturnMapping(None)
    async def Ping(self):
        """Returns -> None"""
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Pinger", request="Ping", version=1, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def Stop(self):
        """Returns -> None"""
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Pinger", request="Stop", version=1, params=_params)

        reply = await self.rpc(msg)
        return reply


class SecretBackendsFacade(Type):
    name = "SecretBackends"
    version = 1

    @ReturnMapping(ErrorResults)
    async def AddSecretBackends(self, args=None):
        """AddSecretBackends adds new secret backends.

        args : typing.Sequence[~AddSecretBackendArg]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="SecretBackends",
            request="AddSecretBackends",
            version=1,
            params=_params,
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ListSecretBackendsResults)
    async def ListSecretBackends(self, names=None, reveal=None):
        """ListSecretBackends lists available secret backends.

        names : typing.Sequence[str]
        reveal : bool
        Returns -> ListSecretBackendsResults
        """
        if names is not None and not isinstance(names, (bytes, str, list)):
            raise Exception(f"Expected names to be a Sequence, received: {type(names)}")

        if reveal is not None and not isinstance(reveal, bool):
            raise Exception(f"Expected reveal to be a bool, received: {type(reveal)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="SecretBackends",
            request="ListSecretBackends",
            version=1,
            params=_params,
        )
        _params["names"] = names
        _params["reveal"] = reveal
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def RemoveSecretBackends(self, args=None):
        """RemoveSecretBackends removes secret backends.

        args : typing.Sequence[~RemoveSecretBackendArg]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="SecretBackends",
            request="RemoveSecretBackends",
            version=1,
            params=_params,
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def UpdateSecretBackends(self, args=None):
        """UpdateSecretBackends updates secret backends.

        args : typing.Sequence[~UpdateSecretBackendArg]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="SecretBackends",
            request="UpdateSecretBackends",
            version=1,
            params=_params,
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply


class SecretsFacade(Type):
    name = "Secrets"
    version = 1

    @ReturnMapping(ListSecretResults)
    async def ListSecrets(self, filter_=None, show_secrets=None):
        """ListSecrets lists available secrets.

        filter_ : SecretsFilter
        show_secrets : bool
        Returns -> ListSecretResults
        """
        if filter_ is not None and not isinstance(filter_, (dict, SecretsFilter)):
            raise Exception(
                f"Expected filter_ to be a SecretsFilter, received: {type(filter_)}"
            )

        if show_secrets is not None and not isinstance(show_secrets, bool):
            raise Exception(
                f"Expected show_secrets to be a bool, received: {type(show_secrets)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Secrets", request="ListSecrets", version=1, params=_params)
        _params["filter"] = filter_
        _params["show-secrets"] = show_secrets
        reply = await self.rpc(msg)
        return reply



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Upgrades previously deployed ubuntu charm

"""

import asyncio

from juju.model import Model


async def main():
    model = Model()
    print("Connecting to model")
    # connect to current model with current user, per Juju CLI
    await model.connect()

    try:
        print("Get deployed application")
        app = model.applications["ubuntu"]

        print("Refresh/Upgrade Ubuntu charm with local charm")
        await app.refresh(path="path/to/local/ubuntu.charm")
    finally:
        print("Disconnecting from model")
        await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2024 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
"""Implementation of Model.wait_for_idle(), analog to `juju wait_for`."""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from typing import AbstractSet

from ..client._definitions import (
    ApplicationStatus,
    FullStatus,
    MachineStatus,
    UnitStatus,
)
from ..errors import JujuAgentError, JujuAppError, JujuMachineError, JujuUnitError

logger = logging.getLogger(__name__)


@dataclass
class CheckStatus:
    """Return type check(), represents single loop iteration."""

    units: set[str]
    """All units visible at this point."""
    ready_units: set[str]
    """Units with the expected workload status."""
    idle_units: set[str]
    """Units with stable (idle) agent status."""


class Loop:
    def __init__(
        self,
        *,
        apps: AbstractSet[str],
        wait_for_exact_units: int | None = None,
        wait_for_units: int,
        idle_period: float,
    ):
        self.apps = apps
        self.wait_for_exact_units = wait_for_exact_units
        self.wait_for_units = wait_for_units
        self.idle_period = idle_period
        self.idle_since: dict[str, float] = {}

    def next(self, status: CheckStatus | None) -> bool:
        logger.info("wait_for_idle iteration %s", status)
        now = time.monotonic()

        if not status:
            return False

        expected_idle_since = now - self.idle_period

        for name in status.units:
            if name in status.idle_units:
                self.idle_since[name] = min(
                    now, self.idle_since.get(name, float("inf"))
                )
            else:
                self.idle_since[name] = float("inf")

        if busy := {n for n, t in self.idle_since.items() if t > expected_idle_since}:
            logger.info("Waiting for units to be idle enough: %s", busy)
            return False

        for app_name in self.apps:
            ready_units = [
                n for n in status.ready_units if n.startswith(f"{app_name}/")
            ]
            if len(ready_units) < self.wait_for_units:
                logger.info(
                    "Waiting for app %r units %s >= %s",
                    app_name,
                    len(status.ready_units),
                    self.wait_for_units,
                )
                return False

            if (
                self.wait_for_exact_units is not None
                and len(ready_units) != self.wait_for_exact_units
            ):
                logger.info(
                    "Waiting for app %r units %s == %s",
                    app_name,
                    len(ready_units),
                    self.wait_for_exact_units,
                )
                return False

        return True


def check(
    full_status: FullStatus,
    *,
    apps: AbstractSet[str],
    raise_on_error: bool,
    raise_on_blocked: bool,
    status: str | None,
) -> CheckStatus | None:
    """A single iteration of a wait_for_idle loop."""
    for app_name in apps:
        if not full_status.applications.get(app_name):
            logger.info("Waiting for app %r", app_name)
            return None

    units: dict[str, UnitStatus] = {}
    rv = CheckStatus(set(), set(), set())

    for app_name in apps:
        units.update(app_units(full_status, app_name))

    if raise_on_error:
        check_errors(full_status, apps, units)

    if raise_on_blocked:
        check_blocked(full_status, apps, units)

    for app_name in apps:
        app = full_status.applications[app_name]
        assert isinstance(app, ApplicationStatus)

        for unit_name, unit in app_units(full_status, app_name).items():
            rv.units.add(unit_name)
            assert unit.agent_status
            assert unit.workload_status

            if unit.agent_status.status == "idle":
                rv.idle_units.add(unit_name)

            if not status or unit.workload_status.status == status:
                rv.ready_units.add(unit_name)

    return rv


def check_errors(
    full_status: FullStatus, apps: AbstractSet[str], units: dict[str, UnitStatus]
) -> None:
    """Check the full status for error conditions, in this order:

    - Machine error (any unit of any app from apps)
    - Agent error (-"-)
    - Workload error (-"-)
    - App error (any app from apps)
    """
    for unit_name, unit in units.items():
        if unit.machine:
            machine = full_status.machines[unit.machine]
            assert isinstance(machine, MachineStatus)
            assert machine.instance_status
            if machine.instance_status.status == "error":
                raise JujuMachineError(
                    f"{unit_name!r} machine {unit.machine!r} has errored: {machine.instance_status.info!r}"
                )

    for unit_name, unit in units.items():
        assert unit.agent_status
        if unit.agent_status.status == "error":
            raise JujuAgentError(
                f"{unit_name!r} agent has errored: {unit.agent_status.info!r}"
            )

    for unit_name, unit in units.items():
        assert unit.workload_status
        if unit.workload_status.status == "error":
            raise JujuUnitError(
                f"{unit_name!r} workload has errored: {unit.workload_status.info!r}"
            )

    for app_name in apps:
        app = full_status.applications[app_name]
        assert isinstance(app, ApplicationStatus)
        assert app.status
        if app.status.status == "error":
            raise JujuAppError(f"{app_name!r} has errored: {app.status.info!r}")


def check_blocked(
    full_status: FullStatus, apps: AbstractSet[str], units: dict[str, UnitStatus]
) -> None:
    """Check the full status for blocked conditions, in this order:

    - Workload blocked (any unit of any app from apps)
    - App blocked (any app from apps)
    """
    for unit_name, unit in units.items():
        assert unit.workload_status
        if unit.workload_status.status == "blocked":
            raise JujuUnitError(
                f"{unit_name!r} workload is blocked: {unit.workload_status.info!r}"
            )

    for app_name in apps:
        app = full_status.applications[app_name]
        assert isinstance(app, ApplicationStatus)
        assert app.status
        if app.status.status == "blocked":
            raise JujuAppError(f"{app_name!r} is blocked: {app.status.info!r}")


def app_units(full_status: FullStatus, app_name: str) -> dict[str, UnitStatus]:
    """Fish out the app's units' status from a FullStatus response."""
    rv: dict[str, UnitStatus] = {}
    app = full_status.applications[app_name]
    assert isinstance(app, ApplicationStatus)

    if app.subordinate_to:
        parent_name = app.subordinate_to[0]
        parent = full_status.applications[parent_name]
        assert isinstance(parent, ApplicationStatus)
        for parent_unit in parent.units.values():
            assert isinstance(parent_unit, UnitStatus)
            for name, unit in parent_unit.subordinates.items():
                if not name.startswith(f"{app_name}/"):
                    continue
                assert isinstance(unit, UnitStatus)
                rv[name] = unit
    else:
        for name, unit in app.units.items():
            assert isinstance(unit, UnitStatus)
            rv[name] = unit

    return rv



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model.
2. Prints out leadership status for all deployed units in the model.
3. Cleanly disconnects.

"""

import asyncio

from juju.model import Model


async def report_leadership():
    model = Model()
    await model.connect()

    print("Leadership: ")
    for app in model.applications.values():
        for unit in app.units:
            print(f"{unit.name}: {await unit.is_leader_from_status()}")

    await model.disconnect()


if __name__ == "__main__":
    asyncio.run(report_leadership())



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class ModelManagerFacade(Type):
    name = "ModelManager"
    version = 9

    @ReturnMapping(ErrorResults)
    async def ChangeModelCredential(self, model_credentials=None):
        """ChangeModelCredential changes cloud credential reference for models.
        These new cloud credentials must already exist on the controller.

        model_credentials : typing.Sequence[~ChangeModelCredentialParams]
        Returns -> ErrorResults
        """
        if model_credentials is not None and not isinstance(
            model_credentials, (bytes, str, list)
        ):
            raise Exception(
                f"Expected model_credentials to be a Sequence, received: {type(model_credentials)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager",
            request="ChangeModelCredential",
            version=9,
            params=_params,
        )
        _params["model-credentials"] = model_credentials
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelInfo)
    async def CreateModel(
        self,
        cloud_tag=None,
        config=None,
        credential=None,
        name=None,
        owner_tag=None,
        region=None,
    ):
        """CreateModel creates a new model using the account and
        model config specified in the args.

        cloud_tag : str
        config : typing.Mapping[str, typing.Any]
        credential : str
        name : str
        owner_tag : str
        region : str
        Returns -> ModelInfo
        """
        if cloud_tag is not None and not isinstance(cloud_tag, (bytes, str)):
            raise Exception(
                f"Expected cloud_tag to be a str, received: {type(cloud_tag)}"
            )

        if config is not None and not isinstance(config, dict):
            raise Exception(
                f"Expected config to be a Mapping, received: {type(config)}"
            )

        if credential is not None and not isinstance(credential, (bytes, str)):
            raise Exception(
                f"Expected credential to be a str, received: {type(credential)}"
            )

        if name is not None and not isinstance(name, (bytes, str)):
            raise Exception(f"Expected name to be a str, received: {type(name)}")

        if owner_tag is not None and not isinstance(owner_tag, (bytes, str)):
            raise Exception(
                f"Expected owner_tag to be a str, received: {type(owner_tag)}"
            )

        if region is not None and not isinstance(region, (bytes, str)):
            raise Exception(f"Expected region to be a str, received: {type(region)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="CreateModel", version=9, params=_params
        )
        _params["cloud-tag"] = cloud_tag
        _params["config"] = config
        _params["credential"] = credential
        _params["name"] = name
        _params["owner-tag"] = owner_tag
        _params["region"] = region
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def DestroyModels(self, models=None):
        """DestroyModels will try to destroy the specified models.
        If there is a block on destruction, this method will return an error.
        From ModelManager v7 onwards, DestroyModels gains 'force' and 'max-wait' parameters.

        models : typing.Sequence[~DestroyModelParams]
        Returns -> ErrorResults
        """
        if models is not None and not isinstance(models, (bytes, str, list)):
            raise Exception(
                f"Expected models to be a Sequence, received: {type(models)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="DestroyModels", version=9, params=_params
        )
        _params["models"] = models
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringResults)
    async def DumpModels(self, entities=None, simplified=None):
        """DumpModels will export the models into the database agnostic
        representation. The user needs to either be a controller admin, or have
        admin privileges on the model itself.

        entities : typing.Sequence[~Entity]
        simplified : bool
        Returns -> StringResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        if simplified is not None and not isinstance(simplified, bool):
            raise Exception(
                f"Expected simplified to be a bool, received: {type(simplified)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelManager", request="DumpModels", version=9, params=_params)
        _params["entities"] = entities
        _params["simplified"] = simplified
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(MapResults)
    async def DumpModelsDB(self, entities=None):
        """DumpModelsDB will gather all documents from all model collections
        for the specified model. The map result contains a map of collection
        names to lists of documents represented as maps.

        entities : typing.Sequence[~Entity]
        Returns -> MapResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="DumpModelsDB", version=9, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelSummaryResults)
    async def ListModelSummaries(self, all_=None, user_tag=None):
        """ListModelSummaries returns models that the specified user
        has access to in the current server.  Controller admins (superuser)
        can list models for any user.  Other users
        can only ask about their own models.

        all_ : bool
        user_tag : str
        Returns -> ModelSummaryResults
        """
        if all_ is not None and not isinstance(all_, bool):
            raise Exception(f"Expected all_ to be a bool, received: {type(all_)}")

        if user_tag is not None and not isinstance(user_tag, (bytes, str)):
            raise Exception(
                f"Expected user_tag to be a str, received: {type(user_tag)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="ListModelSummaries", version=9, params=_params
        )
        _params["all"] = all_
        _params["user-tag"] = user_tag
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(UserModelList)
    async def ListModels(self, tag=None):
        """ListModels returns the models that the specified user
        has access to in the current server.  Controller admins (superuser)
        can list models for any user.  Other users
        can only ask about their own models.

        tag : str
        Returns -> UserModelList
        """
        if tag is not None and not isinstance(tag, (bytes, str)):
            raise Exception(f"Expected tag to be a str, received: {type(tag)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelManager", request="ListModels", version=9, params=_params)
        _params["tag"] = tag
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelDefaultsResults)
    async def ModelDefaultsForClouds(self, entities=None):
        """ModelDefaultsForClouds returns the default config values for the specified
        clouds.

        entities : typing.Sequence[~Entity]
        Returns -> ModelDefaultsResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager",
            request="ModelDefaultsForClouds",
            version=9,
            params=_params,
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelInfoResults)
    async def ModelInfo(self, entities=None):
        """ModelInfo returns information about the specified models.

        entities : typing.Sequence[~Entity]
        Returns -> ModelInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ModelManager", request="ModelInfo", version=9, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelStatusResults)
    async def ModelStatus(self, entities=None):
        """ModelStatus returns a summary of the model.

        entities : typing.Sequence[~Entity]
        Returns -> ModelStatusResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="ModelStatus", version=9, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def ModifyModelAccess(self, changes=None):
        """ModifyModelAccess changes the model access granted to users.

        changes : typing.Sequence[~ModifyModelAccess]
        Returns -> ErrorResults
        """
        if changes is not None and not isinstance(changes, (bytes, str, list)):
            raise Exception(
                f"Expected changes to be a Sequence, received: {type(changes)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="ModifyModelAccess", version=9, params=_params
        )
        _params["changes"] = changes
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def SetModelDefaults(self, config=None):
        """SetModelDefaults writes new values for the specified default model settings.

        config : typing.Sequence[~ModelDefaultValues]
        Returns -> ErrorResults
        """
        if config is not None and not isinstance(config, (bytes, str, list)):
            raise Exception(
                f"Expected config to be a Sequence, received: {type(config)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="SetModelDefaults", version=9, params=_params
        )
        _params["config"] = config
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def UnsetModelDefaults(self, keys=None):
        """UnsetModelDefaults removes the specified default model settings.

        keys : typing.Sequence[~ModelUnsetKeys]
        Returns -> ErrorResults
        """
        if keys is not None and not isinstance(keys, (bytes, str, list)):
            raise Exception(f"Expected keys to be a Sequence, received: {type(keys)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelManager", request="UnsetModelDefaults", version=9, params=_params
        )
        _params["keys"] = keys
        reply = await self.rpc(msg)
        return reply



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class ApplicationOffersFacade(Type):
    name = "ApplicationOffers"
    version = 5

    @ReturnMapping(ApplicationOffersResults)
    async def ApplicationOffers(self, bakery_version=None, offer_urls=None):
        """ApplicationOffers gets details about remote applications that match given URLs.

        bakery_version : int
        offer_urls : typing.Sequence[str]
        Returns -> ApplicationOffersResults
        """
        if bakery_version is not None and not isinstance(bakery_version, int):
            raise Exception(
                f"Expected bakery_version to be a int, received: {type(bakery_version)}"
            )

        if offer_urls is not None and not isinstance(offer_urls, (bytes, str, list)):
            raise Exception(
                f"Expected offer_urls to be a Sequence, received: {type(offer_urls)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="ApplicationOffers",
            version=5,
            params=_params,
        )
        _params["bakery-version"] = bakery_version
        _params["offer-urls"] = offer_urls
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def DestroyOffers(self, force=None, offer_urls=None):
        """DestroyOffers removes the offers specified by the given URLs, forcing if necessary.

        force : bool
        offer_urls : typing.Sequence[str]
        Returns -> ErrorResults
        """
        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if offer_urls is not None and not isinstance(offer_urls, (bytes, str, list)):
            raise Exception(
                f"Expected offer_urls to be a Sequence, received: {type(offer_urls)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers", request="DestroyOffers", version=5, params=_params
        )
        _params["force"] = force
        _params["offer-urls"] = offer_urls
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(QueryApplicationOffersResultsV5)
    async def FindApplicationOffers(self, filters=None):
        """FindApplicationOffers gets details about remote applications that match given filter.

        filters : typing.Sequence[~OfferFilter]
        Returns -> QueryApplicationOffersResultsV5
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="FindApplicationOffers",
            version=5,
            params=_params,
        )
        _params["Filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ConsumeOfferDetailsResults)
    async def GetConsumeDetails(self, offer_urls=None, user_tag=None):
        """GetConsumeDetails returns the details necessary to pass to another model
        to allow the specified args user to consume the offers represented by the args URLs.

        offer_urls : OfferURLs
        user_tag : str
        Returns -> ConsumeOfferDetailsResults
        """
        if offer_urls is not None and not isinstance(offer_urls, (dict, OfferURLs)):
            raise Exception(
                f"Expected offer_urls to be a OfferURLs, received: {type(offer_urls)}"
            )

        if user_tag is not None and not isinstance(user_tag, (bytes, str)):
            raise Exception(
                f"Expected user_tag to be a str, received: {type(user_tag)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="GetConsumeDetails",
            version=5,
            params=_params,
        )
        _params["offer-urls"] = offer_urls
        _params["user-tag"] = user_tag
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(QueryApplicationOffersResultsV5)
    async def ListApplicationOffers(self, filters=None):
        """ListApplicationOffers gets deployed details about application offers that match given filter.
        The results contain details about the deployed applications such as connection count.

        filters : typing.Sequence[~OfferFilter]
        Returns -> QueryApplicationOffersResultsV5
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="ListApplicationOffers",
            version=5,
            params=_params,
        )
        _params["Filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def ModifyOfferAccess(self, changes=None):
        """ModifyOfferAccess changes the application offer access granted to users.

        changes : typing.Sequence[~ModifyOfferAccess]
        Returns -> ErrorResults
        """
        if changes is not None and not isinstance(changes, (bytes, str, list)):
            raise Exception(
                f"Expected changes to be a Sequence, received: {type(changes)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="ModifyOfferAccess",
            version=5,
            params=_params,
        )
        _params["changes"] = changes
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def Offer(self, offers=None):
        """Offer makes application endpoints available for consumption at a specified URL.

        offers : typing.Sequence[~AddApplicationOffer]
        Returns -> ErrorResults
        """
        if offers is not None and not isinstance(offers, (bytes, str, list)):
            raise Exception(
                f"Expected offers to be a Sequence, received: {type(offers)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ApplicationOffers", request="Offer", version=5, params=_params)
        _params["Offers"] = offers
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(RemoteApplicationInfoResults)
    async def RemoteApplicationInfo(self, bakery_version=None, offer_urls=None):
        """RemoteApplicationInfo returns information about the requested remote application.
        This call currently has no client side API, only there for the Dashboard at this stage.

        bakery_version : int
        offer_urls : typing.Sequence[str]
        Returns -> RemoteApplicationInfoResults
        """
        if bakery_version is not None and not isinstance(bakery_version, int):
            raise Exception(
                f"Expected bakery_version to be a int, received: {type(bakery_version)}"
            )

        if offer_urls is not None and not isinstance(offer_urls, (bytes, str, list)):
            raise Exception(
                f"Expected offer_urls to be a Sequence, received: {type(offer_urls)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="RemoteApplicationInfo",
            version=5,
            params=_params,
        )
        _params["bakery-version"] = bakery_version
        _params["offer-urls"] = offer_urls
        reply = await self.rpc(msg)
        return reply


class SubnetsFacade(Type):
    name = "Subnets"
    version = 5

    @ReturnMapping(ZoneResults)
    async def AllZones(self):
        """AllZones returns all availability zones known to Juju. If a
        zone is unusable, unavailable, or deprecated the Available
        field will be false.

        Returns -> ZoneResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Subnets", request="AllZones", version=5, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ListSubnetsResults)
    async def ListSubnets(self, space_tag=None, zone=None):
        """ListSubnets returns the matching subnets after applying
        optional filters.

        space_tag : str
        zone : str
        Returns -> ListSubnetsResults
        """
        if space_tag is not None and not isinstance(space_tag, (bytes, str)):
            raise Exception(
                f"Expected space_tag to be a str, received: {type(space_tag)}"
            )

        if zone is not None and not isinstance(zone, (bytes, str)):
            raise Exception(f"Expected zone to be a str, received: {type(zone)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Subnets", request="ListSubnets", version=5, params=_params)
        _params["space-tag"] = space_tag
        _params["zone"] = zone
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(SubnetsResults)
    async def SubnetsByCIDR(self, cidrs=None):
        """SubnetsByCIDR returns the collection of subnets matching each CIDR in the input.

        cidrs : typing.Sequence[str]
        Returns -> SubnetsResults
        """
        if cidrs is not None and not isinstance(cidrs, (bytes, str, list)):
            raise Exception(f"Expected cidrs to be a Sequence, received: {type(cidrs)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Subnets", request="SubnetsByCIDR", version=5, params=_params)
        _params["cidrs"] = cidrs
        reply = await self.rpc(msg)
        return reply
